/*
** Module   : WidgetPool.java
** Abstract : widget pools management functions
**
** Copyright (c) 2010-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------------------Description---------------------------------
** 001 SVL 20101004 Created initial version with no implementation.
** 002 CA  20130131 Moved to util package (as other non-widget resources can be pooled). Added 
**                  DELETE WIDGET-POOL support.
** 003 CA  20130918 Added runtime support for widget-pools.
*/

package com.goldencode.p2j.util;

import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.goldencode.p2j.security.*;

/**
 * Widget-pool management class. Both type of pools (named or unnamed) follow the scope of a 
 * external program, procedure, function or trigger block. The following rules describe how named
 * and unnamed pools work:
 * <p>
 * <h3>Unnamed Pools</h3>
 * Each created unnamed pool (non-persistent) is active for the current scope.  In the current 
 * scope, unnamed pools are "stacked".  A chain of CREATE WIDGET-POOL statements will create a set
 * of unnamed pools and they will be deleted in the reversed order of the creation, in LIFO order.
 * <p> 
 * Once a persistent unnamed pool is created, a DELETE WIDGET-POOL statement will not be able to
 * delete it, if the persistent pool is on top of the stack.  For this reason, a 
 * CREATE WIDGET-POOL PERSISTENT statement will do nothing if the unnamed pool at the top of the
 * stack is persistent too. 
 * <p> 
 * When the current scope ends, all non-persistent unnamed pools will be deleted. But, if there
 * is an unnamed persistent pool on the top of the stack for scope S2, this one will be "leaked" 
 * to the previous scope, S1.  The reason for this is that if scope S1 has an active unnamed
 * non-persistent pool, then this pool must be deleted when S1 ends.
 * 
 * <h3>Named Pools</h3>
 * At a certain time, no more than one named pool may exist with a given name. When a scope ends,
 * all non-persistent named pools will be deleted. A persistent pool will survive until someone
 * calls a DELETE WIDGET-POOL statement for that persistent pool; this means a persistent pool
 * will survive even after the external procedure which created it is finished or gets deleted.
 * 
 * <h3>Resource management</h3>
 * When a resource (which can be pooled) is created, its IN WIDGET-POOL clause (if present) is
 * validated first. If a named pool is specified and it can't be found/isn't valid, then the 
 * resource will not be created.
 * <p>
 * Adding a resource to an unnamed pool will search for the first active unnamed pool, either in
 * current scope or any previous scope. Once a resource is added to a named or unnamed pool,
 * it will survive until that pool gets deleted (or the resource is deleted).
 * <p>
 * When an external procedure is ran persistent, all resources will survive until the procedure is
 * deleted; when the procedure is deleted, it will delete all the resources in the non-persistent
 * pools.
 */
public class WidgetPool
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(WidgetPool.class);
   
   /** The widget pools are private per each context. */
   private static final ContextLocal<WorkArea> local = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      {
         return new WorkArea();
      };
   };
   
   /**
    * Delete an unnamed pool.
    */
   public static void delete()
   {
      delete((character) null);
   }

   /**
    * Delete a pool with the given name.
    *
    * @param    name
    *           The name of the newly created pool. Specify <code>null</code> is the pool is 
    *           unnamed.
    */
   public static void delete(String name)
   {
      delete(name == null ? (character) null : new character(name));
   }

   /**
    * Delete a pool with the given name.
    *
    * @param    name
    *           The name of the newly created pool. Specify <code>null</code> is the pool is 
    *           unnamed.
    */
   public static void delete(character name)
   {
      if (name != null && name.isUnknown())
      {
         final String errMsg = "Unable to evaluate widget pool name";
         ErrorManager.recordOrThrowError(3177, errMsg, false);
         return;
      }
      
      boolean unnamed = (name == null);
      String sname = (unnamed ? null : name.toStringMessage());
      WidgetPoolData data = getPool(sname, true);
      
      if (data == null)
      {
         if (!unnamed)
         {
            final String msg = "Attempt to delete widget-pool %s, but it doesn't exist";
            ErrorManager.recordOrThrowError(4117, String.format(msg, sname), false);
         }
         
         return;
      }

      deletePool(data);
   }

   /**
    * Create an unnamed pool.
    */
   public static void create()
   {
      create((character) null);
   }

   /**
    * Create an unnamed pool.
    * 
    * @param    persistent
    *           <code>true</code> if the pool is persistent.
    */
   public static void create(boolean persistent)
   {
      create((character) null, persistent);
   }

   /**
    * Create a pool with the given name.
    *
    * @param    name
    *           The name of the newly created pool. Specify <code>null</code> is the pool is 
    *           unnamed.
    */
   public static void create(character name)
   {
      create(name, false);
   }

   /**
    * Create a pool with the given name.
    *
    * @param    name
    *           The name of the newly created pool. Specify <code>null</code> is the pool is 
    *           unnamed.
    */
   public static void create(String name)
   {
      create(name == null ? (character) null : new character(name), false);
   }

   /**
    * Create a pool with the given name and the given "persistent" option.
    *
    * @param    name
    *           The name of the newly created pool. Specify <code>null</code> is the pool is 
    *           unnamed.
    * @param    persistent
    *           <code>true</code> if the pool is persistent.
    */
   public static void create(String name, boolean persistent)
   {
      create(name == null ? (character) null : new character(name), persistent);
   }

   /**
    * Create a pool with the given name and the given "persistent" option.
    *
    * @param    name
    *           The name of the newly created pool. Specify <code>null</code> is  the pool is 
    *           unnamed.
    * @param    persistent
    *           <code>true</code> if the pool is persistent.
    */
   public static void create(character name, boolean persistent)
   {
      if (name != null)
      {
         if (name.isUnknown())
         {
            final String errMsg = "Unable to evaluate widget pool name";
            ErrorManager.recordOrThrowError(3177, errMsg, false);
            return;
         }
         else if (name.toStringMessage().length() == 0)
         {
            final String errMsg = "Attempt to create a widget-pool with a zero-length name";
            ErrorManager.recordOrThrowError(4115, errMsg, false);
            return;
         }
      }

      boolean unnamed = (name == null);
      String sname = (unnamed ? null : name.toStringMessage());
      
      WidgetPoolData data = getPool(sname, true);
      
      if (data != null && !unnamed)
      {
         // error is shown only in case there is already a named pool
         final String msg = "Attempt to create widget-pool %s, but it already exists";
         ErrorManager.recordOrThrowError(4116, String.format(msg, sname), false);

         return;
      }

      // any existing unnamed pools will be hidden in this scope.
      data = newPool(sname, unnamed, persistent);
   }

   /**
    * Adds the specified resource to the pool. Although this is exposed for the public, do not 
    * used outside of P2J code. Is made public because access to it is needed from any P2J 
    * package.
    * 
    * @param    resource
    *           The resource to be pooled.
    * @param    name
    *           The pool name, set to <code>null</code> if the unnamed pool is to be used.
    */
   public static void addResource(WrappedResource resource, character name)
   {
      WorkArea wa = locate();
      
      boolean unnamed = (name == null);
      String sname = (unnamed ? null : name.toStringMessage());
      
      WidgetPoolData data = getPool(sname, false);
      if (data == null)
      {
         if (!unnamed)
         {
            final String msg = 
               "Could not determine named pool '%s'. You need to validate before adding the " +
               "resource!";
            throw new IllegalStateException(String.format(msg, sname));
         }

         // no pool is active, return
         return;
      }

      data.addResource(resource);
      
      // save the resource, to be able to determine its pool later
      wa.poolForResource.put(resource, data.id);
   }
   
   /**
    * Determine if the specified <code>name</name> targets a known widget pool.
    * <p>
    * A missing unnamed pool will be an error.
    * 
    * @param    name
    *           The pool name, set to <code>null</code> if the unnamed pool is to be used.
    *           
    * @return   <code>true</code> if a pool was found with the specified name.
    */
   public static boolean validWidgetPool(character name)
   {
      // an unnamed widget-pool name is always valid.
      if (name == null)
      {
         return true;
      }
      
      if (name.isUnknown())
      {
         final String errMsg = "Unable to evaluate widget pool name";
         ErrorManager.recordOrThrowError(3177, errMsg, false);
         return false;
      }
      
      String sname = name.toStringMessage();
      
      WidgetPoolData data = getPool(sname, false);
      if (data == null)
      {
         // error is shown only if named pool is not found
         final String msg = 
            "Trying to create a widget in widget-pool '%s', but pool does not exist";
         ErrorManager.recordOrThrowError(3178, String.format(msg, sname), false);

         return false;
      }
      
      return true;
   }
   
   /**
    * Notify a new scope has started; invoked by {@link ProcedureManager}'s scoping support when
    * a new scope is started.
    */
   static void scopeStart()
   {
      WorkArea wa = locate();

      wa.namedPoolScopes.push(new HashSet<Integer>());
      wa.unnamedPoolScopes.push(new ArrayDeque<Integer>());
   }
   
   /**
    * Notify a scope has finished; invoked by {@link ProcedureManager}'s scoping support when a
    * scope is finished.
    * <p>
    * This will delete the created non-persistent pools, but only if this is not a notification 
    * of an external procedure which was ran persistent.
    * 
    * @param    oThis
    *           Current THIS-PROCEDURE handle.
    * @param    external
    *           Flag indicating that this is an external procedure.
    * @param    persistent
    *           Flag indicating that this external procedure is ran persistent.
    */
   static void scopeFinished(Object oThis, boolean external, boolean persistent)
   {
      WorkArea wa = locate();
      
      // compute the set of active pools
      Set<Integer> ids = new HashSet<Integer>();
      ids.addAll(wa.namedPoolScopes.peek());
      ids.addAll(wa.unnamedPoolScopes.peek());
      
      if (!(external && persistent))
      {
         deleteNonPersistentPools(ids);
         
         if (external)
         {
            wa.poolsPerExtProg.remove(oThis);
         }
      }
      else
      {
         // if finishing an external program ran in persistent mode, let the pools survive; 
         // they will be deleted when the external program gets deleted
      }

      wa.namedPoolScopes.pop();
      Deque<Integer> unnamedIds = wa.unnamedPoolScopes.pop();
      
      // at the time of the pop, all the non-persistent pools have been removed. now, check if
      // a persistent unnamed pool is in effect. if so, "leak" it to the previous scope.
      if (!unnamedIds.isEmpty())
      {
         // what is left must be a persistent unnamed pool.
         int id = unnamedIds.peek();
         
         if (unnamedIds.size() != 1 || !wa.persistentPools.contains(id))
         {
            // log a warning. pools are not merged in 4GL, as the non-persistent pools will be
            // deleted when the procedure scope ends or the proc handle gets deleted
            if (LOG.isLoggable(Level.INFO))
            {
               final String msg = "More than one active unnamed persistent pool at scope [%d]!";
               LOG.log(Level.INFO, String.format(msg, wa.unnamedPoolScopes.size() + 1));
            }
         }
         
         // leak the pool to the previous scope. if the previous scope has unnamed pools, they 
         // will be hidden by this pool.
         if (!wa.unnamedPoolScopes.isEmpty())
         {
            wa.unnamedPoolScopes.peek().push(id);
         }
      }
   }
   
   /**
    * Delete all the non-persistent widget-pools associated with the specified procedure.
    * 
    * @param    proc
    *           The external procedure instance.
    */
   static void deleteWidgetPools(Object proc)
   {
      // an external procedure was deleted, this acts like it's scope has finished.
      WorkArea wa = locate();
      
      Set<Integer> ids = wa.poolsPerExtProg.get(proc);
      if (ids != null)
      {
         deleteNonPersistentPools(ids);
      }
   }
   
   /**
    * Remove the given resource from its pool, if it was pooled.
    * 
    * @param    resource
    *           The resource to be removed.
    */
   static void removeResource(WrappedResource resource)
   {
      WorkArea wa = locate();

      // nothing to do if the resource wasn't pooled
      if (!wa.poolForResource.containsKey(resource))
      {
         return;
      }

      int poolId = wa.poolForResource.remove(resource);

      WidgetPoolData data = getPool(poolId);
      if (data == null)
      {
         // this resource wasn't pooled, so nothing to do
         return;
      }

      data.removeResource(resource);
   }
   
   /**
    * Locate the context's {@link WorkArea} instance.
    * 
    * @return   See above.
    */
   private static WorkArea locate()
   {
      return local.get();
   }
   
   /***
    * Create a new pool and register it in the appropriate maps/sets.
    * 
    * @param    name
    *           The name of this pool. Is <code>null</code> if the pool is unnamed.
    * @param    unnamed
    *           Flag determining if this pool is unnamed.
    * @param    persistent
    *           Flag identifying if this pool is persistent.
    *           
    * @return   The created {@link WidgetPoolData definition} for the new pool.
    */
   private static WidgetPoolData newPool(String name, boolean unnamed, boolean persistent)
   {
      WorkArea wa = locate();
    
      if (unnamed && persistent)
      {
         Integer top = wa.unnamedPoolScopes.peek().peek();
         if (top != null && wa.persistentPools.contains(top))
         {
            // if the top pool is persistent, then do not create another one. even if 4GL might
            // create a different pool, there is no reason to do so, because we can't pop from
            // the stack a persistent unnamed pool.
            return wa.pools.get(top);
         }
      }
      
      int id = Utils.uniqueId();
      WidgetPoolData data = new WidgetPoolData(id, name);
      wa.pools.put(id, data);

      if (unnamed)
      {
         wa.unnamedPools.add(id);
      }
      else
      {
         wa.namedPools.put(data.key, id);
      }
      
      if (persistent)
      {
         wa.persistentPools.add(id);
      }
      
      if (unnamed)
      {
         wa.unnamedPoolScopes.peek().push(id);
      }
      else
      {
         wa.namedPoolScopes.peek().add(id);
      }
      
      // collect the pools for each external program
      handle hThis = ProcedureManager.thisProcedure();
      Object oThis = hThis.get();
      
      Set<Integer> thisIds = wa.poolsPerExtProg.get(oThis);
      if (thisIds == null)
      {
         thisIds = new HashSet<Integer>();
         wa.poolsPerExtProg.put(oThis, thisIds);
      }
      thisIds.add(id);
      
      // save the procedure which created this pool
      wa.procForPool.put(id, oThis);
      
      return data;
   }

   /**
    * Filter the set of pool IDs and from it, delete only the non-persistent widget-pools.
    * 
    * @param    ids
    *           A set of pool IDs.
    */
   private static void deleteNonPersistentPools(Set<Integer> ids)
   {
      WorkArea wa = locate();
      
      // determine the non-persistent pools
      Set<Integer> nonPersistent = new HashSet<Integer>(ids);
      nonPersistent.removeAll(wa.persistentPools);
      
      // delete the non-persistent pools; persistent pools will survive, even after a procedure 
      // ran non-persistent is finished (or deleted).
      for (int id : nonPersistent)
      {
         WidgetPoolData data = getPool(id);
         
         deletePool(data);
      }
   }
   
   /**
    * Delete the specified pool. Note that an unnamed persistent pool will not be deleted.
    * <p>
    * When a pool is deleted, all its contained resources will be deleted too.
    * 
    * @param    data
    *           The definition of the pool to be deleted.
    */
   private static void deletePool(WidgetPoolData data)
   {
      WorkArea wa = locate();
      int id = data.id;

      boolean unnamed    = wa.unnamedPools.contains(id);
      boolean named      = wa.namedPools.containsKey(data.key);
      boolean persistent = wa.persistentPools.contains(id);
      
      if (unnamed && persistent)
      {
         // the unnamed persistent pool can't be deleted
         return;
      }

      // delete the resources; this will clean the resource-related maps
      data.deleteResources();

      // clean up the registry
      wa.pools.remove(id);
      if (unnamed)
      {
         wa.unnamedPools.remove(id);
      }
      else
      {
         wa.namedPools.remove(data.key);
      }
      wa.persistentPools.remove(id);
      // remove from the external program's set too.
      Object oThis = wa.procForPool.remove(id);
      Set<Integer> ids = wa.poolsPerExtProg.get(oThis);
      if (ids != null)
      {
         // ids may be null if the program which created this pool has been deleted.
         ids.remove(id);
      }

      // the other maps/sets
      if (unnamed)
      {
         wa.unnamedPoolScopes.peek().remove(id);
      }
      else
      {
         // this might be from a previous scope, so need to walk the stack
         for (Set<Integer> namedPools : wa.namedPoolScopes)
         {
            if (namedPools.remove(id))
            {
               // terminate on first found
               break;
            }
         }
      }
   }
   
   /**
    * Determine the active pool. If <code>name</code> is not <code>null</code>, than the active
    * named pool will be used.
    * 
    * @param    name
    *           The pool name. May be <code>null</code> for an unnamed pool.
    * @param    currentScope
    *           <code>true</code> when search must be done only in current scope.
    * @return   <code>null</code> if no pool is found, else its 
    *           {@link WidgetPoolData definition} is returned.
    */
   private static WidgetPoolData getPool(String name, boolean currentScope)
   {
      WorkArea wa = locate();
      
      Integer id = null;
      boolean unnamed = (name == null);
      
      if (!unnamed)
      {
         // get the named pool
         String key = name.toLowerCase();
         id = wa.namedPools.get(key);
      }
      else if (wa.unnamedPools.size() > 0)
      {
         // if no registered unnamed pool, then no reason to seach down the stack

         // get the unnamed pool for this scope; will be null if no unnamed in stack. always
         // check down the stack(s) until one is found
         for (Deque<Integer> stack : wa.unnamedPoolScopes)
         {
            if (!stack.isEmpty())
            {
               id = stack.peek();
               break;
            }
            
            if (currentScope)
            {
               // exit on first scope, if needed
               break;
            }
         }
      }

      return (id == null ? null : getPool(id));
   }
   
   /**
    * Get the {@link WidgetPoolData pool} with the given ID.
    * 
    * @param    id
    *           The pool's ID.
    *           
    * @return   See above.
    */
   private static WidgetPoolData getPool(int id)
   {
      WorkArea wa = locate();
      
      return wa.pools.get(id);
   }

   /**
    * Container for the widget-pool data.
    */
   private static class WorkArea
   {
      /** A registry of pools, per its ID. */
      private final Map<Integer, WidgetPoolData> pools = new HashMap<Integer, WidgetPoolData>();

      /** Pool names are global, so keep a map of them (key is always lowercased). */
      private final Map<String, Integer> namedPools = new HashMap<String, Integer>();
      
      /** A set containing the ID of all unnamed pools. */
      private final Set<Integer> unnamedPools = new HashSet<Integer>();
      
      /** A set containing the ID of all persistent pools. */
      private final Set<Integer> persistentPools = new HashSet<Integer>();

      /** Identify the pool to which a resource was added (if was pooled). */
      private final Map<WrappedResource, Integer> poolForResource = 
         new HashMap<WrappedResource, Integer>();
      
      /** Identify the external procedure used to create the pool. */
      private final Map<Integer, Object> procForPool = new HashMap<Integer, Object>();

      /** Collect all the created pools for an external program. */
      private final Map<Object, Set<Integer>> poolsPerExtProg =
         new HashMap<Object, Set<Integer>>();
      
      /** Keep the scopes of named pools. */
      private final Deque<Set<Integer>> namedPoolScopes = new ArrayDeque<Set<Integer>>();
      
      /** Keep the scopes of unnamed pools (only one can be active at a certain time). */
      private final Deque<Deque<Integer>> unnamedPoolScopes = new ArrayDeque<Deque<Integer>>();
   }
   
   /**
    * The data needed for a widget pool, named or unnamed.
    */
   private static class WidgetPoolData
   {
      /** Set of resources created in this widget pool. */
      private final Set<WrappedResource> resources = new HashSet<WrappedResource>();

      /** The ID of this widget pool. */
      private final int id;
      
      /** This pool's key (to identify it by name, if this is a named pool). */
      private final String key;
      
      /**
       * Create a new widget pool.
       * 
       * @param    id
       *           The ID of this widget pool.
       * @param    name
       *           The name of this widget pool. <code>null</code> if unnamed. 
       */
      public WidgetPoolData(int id, String name)
      {
         this.id = id;
         this.key = (name == null ? null : name.toLowerCase());
      }
      
      /**
       * Add a new resource to this widget pool.
       * 
       * @param    resource
       *           The resource to be added.
       */
      public void addResource(WrappedResource resource)
      {
         resources.add(resource);
      }
      
      /**
       * Remove the resource from this widget pool.
       * 
       * @param    resource
       *           The resource to be removed.
       */
      public void removeResource(WrappedResource resource)
      {
         resources.remove(resource);
      }
      
      /**
       * Delete all the resources from this pool.
       */
      public void deleteResources()
      {
         // use a copy, as the resource-related maps/sets get cleaned when the resource is deleted
         Set<WrappedResource> copy = new HashSet<WrappedResource>(resources);

         for (WrappedResource res : copy)
         {
            HandleOps.delete(new handle(res));
         }
      }
   }
}
